@@ -1,5 +1,7 @@
 module Main (main) where
 
+import Data.Bifunctor (first)
+import Data.ByteArray (constEq)
 import qualified Data.ByteString as BS
 import qualified Data.ByteString.Lazy as BL
 import Data.Maybe (fromMaybe)
@@ -7,13 +9,26 @@ import Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Text.Encoding as TE
 import qualified Network.HTTP.Types as HTTP
-import Network.Wai (Application, Request, Response, ResponseReceived, pathInfo, requestHeaders, requestMethod, responseLBS, strictRequestBody)
+import Network.Wai
+  ( Application,
+    Request,
+    RequestBodyLength (..),
+    Response,
+    ResponseReceived,
+    pathInfo,
+    requestBodyLength,
+    requestHeaders,
+    requestMethod,
+    responseLBS,
+    strictRequestBody,
+  )
 import qualified Network.Wai.Handler.Warp as Warp
 import NovaCache.NarInfo (NarInfo (..), parseNarInfo, renderNarInfo)
 import NovaCache.Signing (SecretKey, parseSecretKey, sign)
-import NovaCache.Store (FileStore, getCacheInfo, newFileStore, readNar, readNarInfo, writeNar, writeNarInfo)
+import NovaCache.Store (FileStore, getCacheInfo, listNarInfoHashes, newFileStore, readNar, readNarInfo, writeNar, writeNarInfo)
 import NovaCache.Validate (validateNarInfo)
 import System.Environment (getArgs, lookupEnv)
+import System.IO (hPutStrLn, stderr)
 import Text.Read (readMaybe)
 
 -- ---------------------------------------------------------------------------
@@ -36,6 +51,10 @@ apiKeyEnvVar = "CACHE_API_KEY"
 signingKeyEnvVar :: String
 signingKeyEnvVar = "SIGNING_KEY_FILE"
 
+-- | Maximum allowed request body size (100 MB).
+maxBodySize :: Int
+maxBodySize = 100 * 1024 * 1024
+
 -- ---------------------------------------------------------------------------
 -- Server configuration
 -- ---------------------------------------------------------------------------
@@ -86,8 +105,8 @@ main = do
 loadSigningKey :: Maybe FilePath -> IO (Maybe SecretKey)
 loadSigningKey Nothing = pure Nothing
 loadSigningKey (Just path) = do
-  contents <- T.strip . TE.decodeUtf8 <$> BS.readFile path
-  case parseSecretKey contents of
+  raw <- BS.readFile path
+  case first show (TE.decodeUtf8' raw) >>= parseSecretKey . T.strip of
     Left err -> do
       putStrLn ("WARNING: failed to load signing key: " ++ err)
       pure Nothing
@@ -103,6 +122,11 @@ app cfg req respond = case (requestMethod req, pathInfo req) of
   -- GET /nix-cache-info
   ("GET", ["nix-cache-info"]) ->
     respond (responseLBS HTTP.status200 textHeaders (BL.fromStrict (renderCacheInfo (cfgStore cfg))))
+  -- GET /narinfo-hashes
+  ("GET", ["narinfo-hashes"]) -> do
+    hashes <- listNarInfoHashes (cfgStore cfg)
+    let body = TE.encodeUtf8 (T.unlines hashes)
+    respond (responseLBS HTTP.status200 textHeaders (BL.fromStrict body))
   -- GET /<hash>.narinfo
   ("GET", [hashNarinfo])
     | Just hashKey <- T.stripSuffix ".narinfo" hashNarinfo -> do
@@ -124,27 +148,67 @@ app cfg req respond = case (requestMethod req, pathInfo req) of
   ("PUT", [hashNarinfo])
     | Just hashKey <- T.stripSuffix ".narinfo" hashNarinfo ->
         requireAuth cfg req respond $ do
-          body <- BL.toStrict <$> strictRequestBody req
-          case parseNarInfo (TE.decodeUtf8 body) of
-            Left err ->
-              respond (badRequest (T.pack err))
-            Right ni -> case validateNarInfo ni of
-              Left errs ->
-                respond (badRequest (T.unlines (map (T.pack . show) errs)))
-              Right _ -> do
-                let signed = signNarInfo (cfgSigningKey cfg) body
-                writeNarInfo (cfgStore cfg) hashKey signed
-                respond (responseLBS HTTP.status200 textHeaders "ok")
+          bodyResult <- readBodyLimited req
+          case bodyResult of
+            Nothing ->
+              respond (responseLBS HTTP.status413 textHeaders "request body too large")
+            Just body -> case decodeAndValidate body of
+              Left err ->
+                respond (badRequest err)
+              Right ni -> do
+                signed <- signNarInfo (cfgSigningKey cfg) ni
+                ok <- writeNarInfo (cfgStore cfg) hashKey signed
+                if ok
+                  then respond (responseLBS HTTP.status200 textHeaders "ok")
+                  else respond (badRequest "invalid path")
   -- PUT /nar/<file> (auth required)
   ("PUT", ["nar", fileName]) ->
     requireAuth cfg req respond $ do
-      body <- BL.toStrict <$> strictRequestBody req
-      writeNar (cfgStore cfg) fileName body
-      respond (responseLBS HTTP.status200 textHeaders "ok")
+      bodyResult <- readBodyLimited req
+      case bodyResult of
+        Nothing ->
+          respond (responseLBS HTTP.status413 textHeaders "request body too large")
+        Just body -> do
+          ok <- writeNar (cfgStore cfg) fileName body
+          if ok
+            then respond (responseLBS HTTP.status200 textHeaders "ok")
+            else respond (badRequest "invalid path")
   -- Fallback
   _ ->
     respond (responseLBS HTTP.status404 textHeaders "not found")
 
+-- ---------------------------------------------------------------------------
+-- Validation pipeline
+-- ---------------------------------------------------------------------------
+
+-- | Decode a raw narinfo body and validate it in a single pure pipeline.
+--
+-- Composes UTF-8 decoding, narinfo parsing, and field validation. Returns
+-- the validated 'NarInfo' on success, or a user-facing error message.
+decodeAndValidate :: BS.ByteString -> Either Text NarInfo
+decodeAndValidate body = do
+  decoded <- first (const "request body is not valid UTF-8") (TE.decodeUtf8' body)
+  ni <- first T.pack (parseNarInfo decoded)
+  first (T.unlines . map (T.pack . show)) (validateNarInfo ni)
+
+-- ---------------------------------------------------------------------------
+-- Request body limiting
+-- ---------------------------------------------------------------------------
+
+-- | Read the request body, rejecting payloads over 'maxBodySize'.
+--
+-- Returns 'Nothing' if the declared @Content-Length@ exceeds the limit.
+-- For chunked transfers the body is read and checked after the fact.
+readBodyLimited :: Request -> IO (Maybe BS.ByteString)
+readBodyLimited req = case requestBodyLength req of
+  KnownLength len
+    | len > fromIntegral maxBodySize -> pure Nothing
+  _ -> do
+    body <- BL.toStrict <$> strictRequestBody req
+    if BS.length body > maxBodySize
+      then pure Nothing
+      else pure (Just body)
+
 -- ---------------------------------------------------------------------------
 -- Auth
 -- ---------------------------------------------------------------------------
@@ -153,12 +217,14 @@ app cfg req respond = case (requestMethod req, pathInfo req) of
 --
 -- If no key is configured, all writes are permitted (open mode).
 -- Otherwise the request must carry @Authorization: Bearer \<key\>@.
+-- Uses constant-time comparison to prevent timing attacks.
 requireAuth :: Config -> Request -> (Response -> IO ResponseReceived) -> IO ResponseReceived -> IO ResponseReceived
 requireAuth cfg req respond action = case cfgApiKey cfg of
   Nothing -> action
   Just expected ->
     let provided = lookup HTTP.hAuthorization (requestHeaders req)
-     in if provided == Just ("Bearer " <> expected)
+        expectedHeader = "Bearer " <> expected
+     in if maybe False (constEq expectedHeader) provided
           then action
           else respond (responseLBS HTTP.status401 textHeaders "unauthorized")
 
@@ -166,16 +232,23 @@ requireAuth cfg req respond action = case cfgApiKey cfg of
 -- Signing
 -- ---------------------------------------------------------------------------
 
--- | Sign a narinfo body if a signing key is configured.
-signNarInfo :: Maybe SecretKey -> BS.ByteString -> BS.ByteString
-signNarInfo Nothing body = body
-signNarInfo (Just sk) body = case parseNarInfo (TE.decodeUtf8 body) of
-  Left _ -> body
-  Right ni -> case sign sk ni of
-    Left _ -> body
-    Right sig ->
-      let signed = ni {niSigs = niSigs ni ++ [sig]}
-       in TE.encodeUtf8 (renderNarInfo signed)
+-- | Sign a validated 'NarInfo' if a signing key is configured.
+--
+-- Logs a warning to stderr if signing fails, then returns the unsigned
+-- rendering so the upload is not rejected.
+signNarInfo :: Maybe SecretKey -> NarInfo -> IO BS.ByteString
+signNarInfo Nothing ni = pure (renderNarInfoBytes ni)
+signNarInfo (Just sk) ni = case sign sk ni of
+  Left err -> do
+    hPutStrLn stderr ("WARNING: signNarInfo: sign failed: " ++ err)
+    pure (renderNarInfoBytes ni)
+  Right sig ->
+    let signed = ni {niSigs = niSigs ni ++ [sig]}
+     in pure (renderNarInfoBytes signed)
+
+-- | Render a 'NarInfo' to its UTF-8 encoded wire format.
+renderNarInfoBytes :: NarInfo -> BS.ByteString
+renderNarInfoBytes = TE.encodeUtf8 . renderNarInfo
 
 -- ---------------------------------------------------------------------------
 -- Response helpers
